@@ -55,7 +55,7 @@ public:
   }
   uint32 elements() const
   {
-    return length() / sizeof(MY_XPATH_FLT);
+    return length() / (uint32) sizeof(MY_XPATH_FLT);
   }
 };
 
@@ -110,15 +110,15 @@ protected:
 public:
   Item_xml_str_func(THD *thd, Item *a, Item *b): Item_str_func(thd, a, b)
   {
-    maybe_null= TRUE;
+    set_maybe_null();
   }
   Item_xml_str_func(THD *thd, Item *a, Item *b, Item *c):
     Item_str_func(thd, a, b, c)
   {
-    maybe_null= TRUE;
+    set_maybe_null();
   }
   bool fix_fields(THD *thd, Item **ref) override;
-  bool fix_length_and_dec() override;
+  bool fix_length_and_dec(THD *thd) override;
   bool const_item() const override
   {
     return const_item_cache && (!nodeset_func || nodeset_func->const_item());
@@ -131,9 +131,15 @@ class Item_func_xml_extractvalue: public Item_xml_str_func
 public:
   Item_func_xml_extractvalue(THD *thd, Item *a, Item *b):
     Item_xml_str_func(thd, a, b) {}
-  const char *func_name() const override { return "extractvalue"; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("extractvalue") };
+    return name;
+  }
   String *val_str(String *) override;
-  Item *do_get_copy(THD *thd) const override
+
+protected:
+  Item *shallow_copy(THD *thd) const override
   { return get_item_copy<Item_func_xml_extractvalue>(thd, this); }
 };
 
@@ -148,9 +154,15 @@ class Item_func_xml_update: public Item_xml_str_func
 public:
   Item_func_xml_update(THD *thd, Item *a, Item *b, Item *c):
     Item_xml_str_func(thd, a, b, c) {}
-  const char *func_name() const override { return "updatexml"; }
+  LEX_CSTRING func_name_cstring() const override
+  {
+    static LEX_CSTRING name= {STRING_WITH_LEN("updatexml") };
+    return name;
+  }
   String *val_str(String *) override;
-  Item *do_get_copy(THD *thd) const override
+
+protected:
+  Item *shallow_copy(THD *thd) const override
   { return get_item_copy<Item_func_xml_update>(thd, this); }
 };
 
