@@ -43,7 +43,7 @@
 
   The execution of the transformed query (Q1R) follows these steps:
     1. For each row of t1 where t1.b < const a temporary table
-       containing all rows of of t2 with t2.a = t1.a is created
+       containing all rows of t2 with t2.a = t1.a is created
     2. If there are any rows in the temporary table aggregation
        is performed for them
     3. The result of the aggregation is joined with t1.
@@ -155,7 +155,7 @@
   subsets the operation can applied to each subset independently. In this case
   all rows are first partitioned into the groups each of which contains all the
   rows from the partitions belonging the same subset and then each group
-  is subpartitioned into groups in the the post join operation.
+  is subpartitioned into groups in the post join operation.
 
   The set of all rows belonging to the union of several partitions is called
   here superpartition. If a grouping operation is defined by the list
@@ -229,6 +229,7 @@
 #include "mariadb.h"
 #include "sql_select.h"
 #include "opt_trace.h"
+#include "optimizer_defaults.h"
 
 /* Info on a splitting field */
 struct SplM_field_info
@@ -286,11 +287,10 @@ public:
   List<SplM_plan_info> plan_cache;
   /* Cost of best execution plan for join when nothing is pushed */
   double unsplit_cost;
+  /* Split operation cost (result form spl_postjoin_oper_cost()) */
+  double unsplit_oper_cost;
   /* Cardinality of T when nothing is pushed */
   double unsplit_card;
-  /* Lastly evaluated execution plan for 'join' with pushed equalities */
-  SplM_plan_info *last_plan;
-  double last_refills;
 
   SplM_plan_info *find_plan(TABLE *table, uint key, uint parts);
 };
@@ -355,6 +355,7 @@ struct SplM_field_ext_info: public SplM_field_info
        with available statistics.
     10. The select doesn't use WITH ROLLUP (This limitation can probably be
        lifted)
+    11. The select doesn't have ORDER BY with LIMIT
 
   @retval
     true   if the answer is positive
@@ -388,6 +389,9 @@ bool JOIN::check_for_splittable_materialized()
   if (!partition_list)
     return false;
 
+  if (select_lex->order_list.elements > 0 && !unit->lim.is_unlimited()) //!(11)
+    return false;
+
   Json_writer_object trace_wrapper(thd);
   Json_writer_object trace_split(thd, "check_split_materialized");
 
@@ -509,9 +513,7 @@ bool JOIN::check_for_splittable_materialized()
     the collected info on potential splittability of T
   */
   SplM_opt_info *spl_opt_info= new (thd->mem_root) SplM_opt_info();
-  SplM_field_info *spl_field=
-    (SplM_field_info *) (thd->calloc(sizeof(SplM_field_info) *
-                                            spl_field_cnt));
+  SplM_field_info *spl_field= thd->calloc<SplM_field_info>(spl_field_cnt);
 
   if (!(spl_opt_info && spl_field)) // consider T as not good for splitting
     return false;
@@ -602,20 +604,30 @@ void TABLE::add_splitting_info_for_key_field(KEY_FIELD *key_field)
     condition of the select that specifies this table.
   */
   THD *thd= in_use;
-  Item *left_item= spl_field->producing_item->build_clone(thd);
-  Item *right_item= key_field->val->build_clone(thd);
-  Item_func_eq *eq_item= 0;
+  Item *left_item= spl_field->producing_item->deep_copy_with_checks(thd);
+  Item *right_item= key_field->val->deep_copy_with_checks(thd);
+  Item_bool_func *eq_item= 0;
   if (left_item && right_item)
   {
     right_item->walk(&Item::set_fields_as_dependent_processor,
                      false, join->select_lex);
     right_item->update_used_tables();
-    eq_item= new (thd->mem_root) Item_func_eq(thd, left_item, right_item);
+    /*
+      We've just pushed right_item down into the child select. It may only
+      have references to outside.
+    */
+    DBUG_ASSERT(!(right_item->used_tables() & ~PSEUDO_TABLE_BITS));
+
+    //  Item_func::EQUAL_FUNC is null-safe, others can use Item_func_eq()
+    if (key_field->cond->type() == Item::FUNC_ITEM &&
+        ((Item_func*)key_field->cond)->functype() == Item_func::EQUAL_FUNC)
+      eq_item= new (thd->mem_root) Item_func_equal(thd, left_item, right_item);
+    else
+      eq_item= new (thd->mem_root) Item_func_eq(thd, left_item, right_item);
   }
   if (!eq_item)
     return;
-  KEY_FIELD *added_key_field=
-    (KEY_FIELD *) thd->alloc(sizeof(KEY_FIELD));
+  KEY_FIELD *added_key_field= thd->alloc<KEY_FIELD>(1);
   if (!added_key_field ||
       spl_opt_info->added_key_fields.push_back(added_key_field,thd->mem_root))
     return;
@@ -625,14 +637,7 @@ void TABLE::add_splitting_info_for_key_field(KEY_FIELD *key_field)
   added_key_field->level= 0;
   added_key_field->optimize= KEY_OPTIMIZE_EQ;
   added_key_field->eq_func= true;
-
-  Item *real= key_field->val->real_item();
-  if ((real->type() == Item::FIELD_ITEM) &&
-        ((Item_field*)real)->field->maybe_null())
-    added_key_field->null_rejecting= true;
-  else
-    added_key_field->null_rejecting= false;
-
+  added_key_field->null_rejecting= key_field->null_rejecting;
   added_key_field->cond_guard= NULL;
   added_key_field->sj_pred_no= UINT_MAX;
   return;
@@ -651,7 +656,7 @@ add_ext_keyuse_for_splitting(Dynamic_array<KEYUSE_EXT> *ext_keyuses,
   possible_keys.intersect(field->table->keys_usable_for_splitting);
   tab->keys.merge(possible_keys);
 
-  Item_func_eq *eq_item= (Item_func_eq *) (added_key_field->cond);
+  Item_args *eq_item= (Item_args *) (added_key_field->cond);
   keyuse_ext.table= field->table;
   keyuse_ext.val= eq_item->arguments()[1];
   keyuse_ext.key= key;
@@ -665,7 +670,7 @@ add_ext_keyuse_for_splitting(Dynamic_array<KEYUSE_EXT> *ext_keyuses,
   keyuse_ext.sj_pred_no= added_key_field->sj_pred_no;
   keyuse_ext.validity_ref= 0;
   keyuse_ext.needed_in_prefix= added_key_field->val->used_tables() &
-			       ~(OUTER_REF_TABLE_BIT | RAND_TABLE_BIT);
+                              ~(OUTER_REF_TABLE_BIT | RAND_TABLE_BIT);
   keyuse_ext.validity_var= false;
   return ext_keyuses->push(keyuse_ext);
 }
@@ -726,20 +731,28 @@ add_ext_keyuses_for_splitting_field(Dynamic_array<KEYUSE_EXT> *ext_keyuses,
 /*
   @brief
     Cost of the post join operation used in specification of splittable table
+    This does not include the cost of creating the temporary table as this
+    operation can be executed many times for the same temporary table.
 */
 
 static
 double spl_postjoin_oper_cost(THD *thd, double join_record_count, uint rec_len)
 {
   double cost;
-  cost=  get_tmp_table_write_cost(thd, join_record_count,rec_len) *
-         join_record_count;   // cost to fill tmp table
-  cost+= get_tmp_table_lookup_cost(thd, join_record_count,rec_len) *
-         join_record_count;   // cost to perform post join operation used here
-  cost+= get_tmp_table_lookup_cost(thd, join_record_count, rec_len) +
-         (join_record_count == 0 ? 0 :
-          join_record_count * log2 (join_record_count)) *
-         SORT_INDEX_CMP_COST;             // cost to perform  sorting
+  TMPTABLE_COSTS tmp_cost= get_tmp_table_costs(thd, join_record_count,
+                                               rec_len, 0, 1);
+  /* cost to fill tmp table */
+  cost= tmp_cost.write * join_record_count;
+  /* cost to perform post join operation used here */
+  cost+= tmp_cost.lookup * join_record_count;
+  /* cost to preform sorting */
+  /* QQQ
+     We should use cost_of_filesort() for computing sort.
+     Do we always preform sorting ? If not, this should be done conditionally
+  */
+  cost+= ((join_record_count == 0 ? 0 :
+           join_record_count * log2 (join_record_count)) *
+          SORT_INDEX_CMP_COST);
   return cost;
 }
 
@@ -768,10 +781,9 @@ double spl_postjoin_oper_cost(THD *thd, double join_record_count, uint rec_len)
 void JOIN::add_keyuses_for_splitting()
 {
   uint i;
-  uint idx;
+  size_t idx;
   KEYUSE_EXT *keyuse_ext;
   KEYUSE_EXT keyuse_ext_end;
-  double oper_cost;
   uint rec_len;
   uint added_keyuse_count;
   TABLE *table= select_lex->master_unit()->derived->table;
@@ -794,14 +806,20 @@ void JOIN::add_keyuses_for_splitting()
   if (ext_keyuses_for_splitting->push(keyuse_ext_end))
     goto err;
   // psergey-todo: trace anything here?
-  spl_opt_info->unsplit_card= join_record_count;
+  /*
+    Use the number of rows that was computed by
+    TABLE_LIST::fetch_number_of_rows():
+  */
+  spl_opt_info->unsplit_card=
+    rows2double(select_lex->master_unit()->derived->table->stat_records());
 
   rec_len= table->s->rec_buff_length;
 
-  oper_cost= spl_postjoin_oper_cost(thd, join_record_count, rec_len);
-
-  spl_opt_info->unsplit_cost= best_positions[table_count-1].read_time +
-                              oper_cost;
+  spl_opt_info->unsplit_oper_cost= spl_postjoin_oper_cost(thd,
+                                                          join_record_count,
+                                                          rec_len);
+  spl_opt_info->unsplit_cost= (best_positions[table_count-1].read_time +
+                               spl_opt_info->unsplit_oper_cost);
 
   if (!(save_qep= new Join_plan_state(table_count + 1)))
     goto err;
@@ -830,7 +848,7 @@ void JOIN::add_keyuses_for_splitting()
     added_keyuse->validity_ref= &keyuse_ext->validity_var;
   }
 
-  if (sort_and_filter_keyuse(thd, &keyuse, true))
+  if (sort_and_filter_keyuse(this, &keyuse, true))
     goto err;
   optimize_keyuse(this, &keyuse);
 
@@ -965,6 +983,7 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
   uint best_key= 0;
   uint best_key_parts= 0;
   table_map best_param_tables= 0L;
+  bool chosen, already_printed;
   Json_writer_object trace_obj(thd, "choose_best_splitting");
   Json_writer_array  trace_arr(thd, "considered_keys");
   /*
@@ -1035,13 +1054,32 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
     while (keyuse_ext->table == table);
   }
   trace_arr.end();
+  chosen= 0;
 
-  spl_opt_info->last_plan= 0;
   double refills= DBL_MAX;
   table_map excluded_tables= remaining_tables | this->join->sjm_lookup_tables;
   if (best_table)
   {
     *spl_pd_boundary= this->table->map;
+    /*
+      Compute "refills" - how many times we'll need to refill the split-
+      materialized temp. table. Split-materialized table has references to
+      preceding table(s). Suppose the join prefix is (t1, t2, t3) and
+      split-materialized refers to table t2:
+
+        t1  t2  t3  <split_materialized>
+            ^             |
+            +------------ +
+
+      If we do not use join buffer for table t3, then we'll need to refill
+      the split-materialized table partial_join_cardinality({t1, t2}) times.
+      (this assumes that fanout of table t3 is greater than 1, which is
+      typically true).
+      If table t3 uses join buffer, then every time we get a record combination
+      of {t1.row,t2.row,t3.row} the t2.row may be different and so we will need
+      to refill <split_materialized> every time, that is,
+      partial_join_cardinality(t1,t3,t3) times.
+    */
     if (!best_param_tables)
       refills= 1;
     else
@@ -1080,7 +1118,7 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
       table_map all_table_map= (((table_map) 1) << join->table_count) - 1;
       reset_validity_vars_for_keyuses(best_key_keyuse_ext_start, best_table,
                                       best_key, excluded_tables, true);
-      choose_plan(join, all_table_map & ~join->const_table_map);
+      choose_plan(join, all_table_map & ~join->const_table_map, 0);
 
       wrapper.end();
       /*
@@ -1093,9 +1131,8 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
       key_map spl_keys= table->keys_usable_for_splitting;
       if (!(first_non_const_pos->key &&
             spl_keys.is_set(first_non_const_pos->key->key)) ||
-          !(spl_plan= (SplM_plan_info *) thd->alloc(sizeof(SplM_plan_info))) ||
-	  !(spl_plan->best_positions=
-	     (POSITION *) thd->alloc(sizeof(POSITION) * join->table_count)) ||
+          !(spl_plan= thd->alloc<SplM_plan_info>(1)) ||
+	  !(spl_plan->best_positions= thd->alloc<POSITION>(join->table_count)) ||
 	  spl_opt_info->plan_cache.push_back(spl_plan))
       {
         reset_validity_vars_for_keyuses(best_key_keyuse_ext_start, best_table,
@@ -1113,61 +1150,74 @@ SplM_plan_info * JOIN_TAB::choose_best_splitting(uint idx,
                             spl_opt_info->unsplit_card : 1); 
 
       uint rec_len= table->s->rec_buff_length;
-
       double split_card= spl_opt_info->unsplit_card * spl_plan->split_sel;
-      double oper_cost= split_card *
-                        spl_postjoin_oper_cost(thd, split_card, rec_len);
-      spl_plan->cost= join->best_positions[join->table_count-1].read_time +
-                      + oper_cost;
+      double oper_cost= (split_card *
+                         spl_postjoin_oper_cost(thd, split_card, rec_len));
+      spl_plan->cost= (join->best_positions[join->table_count-1].read_time +
+                       oper_cost);
 
+      chosen= (refills * spl_plan->cost + COST_EPS <
+               spl_opt_info->unsplit_cost);
+
+      if (unlikely(thd->trace_started()))
+      {
+        //psergey-merge:Json_writer_object wrapper(thd);
+        Json_writer_object find_trace(thd, "split_materialized");
+        find_trace.
+          add("table", best_table->alias.c_ptr()).
+          add("key", best_table->key_info[best_key].name).
+          add("org_cost",join->best_positions[join->table_count-1].read_time).
+          add("postjoin_cost", oper_cost).
+          add("one_splitting_cost", spl_plan->cost).
+          add("unsplit_postjoin_cost", spl_opt_info->unsplit_oper_cost).
+          add("unsplit_cost", spl_opt_info->unsplit_cost).
+          add("rows", split_card).
+          add("refills", refills).
+          add("total_splitting_cost", refills * spl_plan->cost).
+          add("chosen", chosen);
+      }
       memcpy((char *) spl_plan->best_positions,
              (char *) join->best_positions,
              sizeof(POSITION) * join->table_count);
       reset_validity_vars_for_keyuses(best_key_keyuse_ext_start, best_table,
                                       best_key, excluded_tables, false);
+      already_printed= 1;
     }
     else
-      trace_obj.add("cached_plan_found", 1);
-
-    if (spl_plan)
     {
-      trace_obj.
-        add("lead_table", spl_plan->table->alias.ptr()).
-        add("index",      spl_plan->table->key_info[spl_plan->key].name).
-        add("parts",      spl_plan->parts).
-        add("split_sel",  spl_plan->split_sel).
-        add("cost",       spl_plan->cost).
-        add("records",    (ha_rows) (records * spl_plan->split_sel)).
-        add("refills",    refills);
-
-      if (refills * spl_plan->cost < spl_opt_info->unsplit_cost - 0.01)
-      {
-        /*
-          The best plan that employs splitting is cheaper than
-          the plan without splitting
-	*/
-        spl_opt_info->last_plan= spl_plan;
-        spl_opt_info->last_refills= refills;
-        trace_obj.add("chosen", true);
-      }
-      else
-        trace_obj.add("chosen", false);
+      trace_obj.add("cached_plan_found", 1);
+      chosen= (refills * spl_plan->cost + COST_EPS <
+               spl_opt_info->unsplit_cost);
+      already_printed= 0;
     }
   }
 
   /* Set the cost of the preferred materialization for this partial join */
-  records= (ha_rows)spl_opt_info->unsplit_card;
-  spl_plan= spl_opt_info->last_plan;
-  if (spl_plan)
+  if (chosen)
   {
-    startup_cost= spl_opt_info->last_refills * spl_plan->cost;
-    records= (ha_rows) (records * spl_plan->split_sel);
+    /*
+      The best plan that employs splitting is cheaper than
+      the plan without splitting
+    */
+    startup_cost= refills * spl_plan->cost;
+    records= (ha_rows) (spl_opt_info->unsplit_card * spl_plan->split_sel);
+    if (unlikely(thd->trace_started()) && ! already_printed)
+    {
+      Json_writer_object trace(thd, "split_materialized");
+      trace.
+        add("one_splitting_cost", spl_plan->cost).
+        add("total_splitting_cost", startup_cost).
+        add("rows", records);
+    }
   }
   else
   {
-    startup_cost= spl_opt_info->unsplit_cost;
-    *spl_pd_boundary= 0;
+    /* Restore original values */
+    startup_cost=      spl_opt_info->unsplit_cost;
+    records= (ha_rows) spl_opt_info->unsplit_card;
+    spl_plan= 0;
   }
+
   return spl_plan;
 }
 
@@ -1236,7 +1286,7 @@ bool JOIN::inject_best_splitting_cond(table_map excluded_tables)
     Test if equality is injected for split optimization
 
   @param
-    eq_item   equality to to test
+    eq_item   equality to test
 
   @retval
     true    eq_item is equality injected for split optimization
