@@ -1,5 +1,5 @@
 /* Copyright (c) 2000, 2010, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2013 Monty Program Ab.
+   Copyright (c) 2009, 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,13 +12,12 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA */
+   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1335  USA */
 
 
 /* Functions to handle date and time */
 
 #include "mariadb.h"
-#include "sql_priv.h"
 #include "sql_time.h"
 #include "tztime.h"                      // struct Time_zone
 #include "sql_class.h"                   // THD
@@ -69,62 +68,71 @@ int append_interval(String *str, interval_type int_type, const INTERVAL &interva
   size_t len;
   switch (int_type) {
   case INTERVAL_YEAR:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.year);
+    len= my_snprintf(buf,sizeof(buf),"%lu", interval.year);
     break;
   case INTERVAL_QUARTER:
   case INTERVAL_MONTH:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.month);
+    len= my_snprintf(buf,sizeof(buf),"%lu", interval.month);
     int_type=INTERVAL_MONTH;
     break;
   case INTERVAL_WEEK:
   case INTERVAL_DAY:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.day);
+    len= my_snprintf(buf,sizeof(buf),"%lu", interval.day);
     int_type=INTERVAL_DAY;
     break;
   case INTERVAL_HOUR:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.hour);
+    len= my_snprintf(buf,sizeof(buf),"%lu", interval.hour);
     break;
   case INTERVAL_MINUTE:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.minute);
+    len= my_snprintf(buf,sizeof(buf),"%llu", interval.minute);
     break;
   case INTERVAL_SECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.second);
+    len= my_snprintf(buf,sizeof(buf),"%llu", interval.second);
     break;
   case INTERVAL_MICROSECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u", interval.second_part);
+    len= my_snprintf(buf,sizeof(buf),"%llu", interval.second_part);
     break;
   case INTERVAL_YEAR_MONTH:
-    len= my_snprintf(buf,sizeof(buf),"%u-%02u", interval.day, interval.month);
+    len= my_snprintf(buf,sizeof(buf),"'%lu-%02lu'",
+                     interval.year, interval.month);
     break;
   case INTERVAL_DAY_HOUR:
-    len= my_snprintf(buf,sizeof(buf),"%u %u", interval.day, interval.hour);
+    len= my_snprintf(buf,sizeof(buf),"'%lu %lu'", interval.day, interval.hour);
     break;
   case INTERVAL_DAY_MINUTE:
-    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u", interval.day, interval.hour, interval.minute);
+    len= my_snprintf(buf,sizeof(buf),"'%lu %lu:%02llu'",
+                     interval.day, interval.hour, interval.minute);
     break;
   case INTERVAL_DAY_SECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u:%02u", interval.day, interval.hour, interval.minute, interval.second);
+    len= my_snprintf(buf,sizeof(buf),"'%lu %lu:%02llu:%02llu'",
+                     interval.day, interval.hour, interval.minute, interval.second);
     break;
   case INTERVAL_HOUR_MINUTE:
-    len= my_snprintf(buf,sizeof(buf),"%u:%02u", interval.hour, interval.minute);
+    len= my_snprintf(buf,sizeof(buf),"'%lu:%02llu'", interval.hour, interval.minute);
     break;
   case INTERVAL_HOUR_SECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u:%02u:%02u", interval.hour, interval.minute, interval.second);
+    len= my_snprintf(buf,sizeof(buf),"'%lu:%02llu:%02llu'",
+                     interval.hour, interval.minute, interval.second);
     break;
   case INTERVAL_MINUTE_SECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u:%02u", interval.minute, interval.second);
+    len= my_snprintf(buf,sizeof(buf),"'%llu:%02llu'", interval.minute, interval.second);
     break;
   case INTERVAL_DAY_MICROSECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u %u:%02u:%02u.%06u", interval.day, interval.hour, interval.minute, interval.second, interval.second_part);
+    len= my_snprintf(buf,sizeof(buf),"'%lu %lu:%02llu:%02llu.%06llu'",
+                     interval.day, interval.hour, interval.minute,
+                     interval.second, interval.second_part);
     break;
   case INTERVAL_HOUR_MICROSECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u:%02u:%02u.%06u", interval.hour, interval.minute, interval.second, interval.second_part);
+    len= my_snprintf(buf,sizeof(buf),"'%lu:%02llu:%02llu.%06llu'",
+                     interval.hour, interval.minute, interval.second,
+                     interval.second_part);
     break;
   case INTERVAL_MINUTE_MICROSECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u:%02u.%06u", interval.minute, interval.second, interval.second_part);
+    len= my_snprintf(buf,sizeof(buf),"'%llu:%02llu.%06llu'",
+                     interval.minute, interval.second, interval.second_part);
     break;
   case INTERVAL_SECOND_MICROSECOND:
-    len= my_snprintf(buf,sizeof(buf),"%u.%06u", interval.second, interval.second_part);
+    len= my_snprintf(buf,sizeof(buf),"%llu.%06llu", interval.second, interval.second_part);
     break;
   default:
     DBUG_ASSERT(0);
@@ -152,7 +160,7 @@ int calc_weekday(long daynr,bool sunday_first_day_of_week)
       		   	  If set	Monday is first day of week
    WEEK_YEAR (1)	  If not set	Week is in range 0-53
 
-   	Week 0 is returned for the the last week of the previous year (for
+   	Week 0 is returned for the last week of the previous year (for
 	a date at start of january) In this case one can get 53 for the
 	first week of next year.  This flag ensures that the week is
 	relevant for the given year. Note that this flag is only
@@ -175,7 +183,7 @@ int calc_weekday(long daynr,bool sunday_first_day_of_week)
 	next week is week 1.
 */
 
-uint calc_week(MYSQL_TIME *l_time, uint week_behaviour, uint *year)
+uint calc_week(const MYSQL_TIME *l_time, uint week_behaviour, uint *year)
 {
   uint days;
   ulong daynr=calc_daynr(l_time->year,l_time->month,l_time->day);
@@ -289,15 +297,15 @@ ulong convert_month_to_period(ulong month)
 
 
 bool
-check_date_with_warn(const MYSQL_TIME *ltime, ulonglong fuzzy_date,
-                     timestamp_type ts_type)
+check_date_with_warn(THD *thd, const MYSQL_TIME *ltime,
+                     date_conv_mode_t fuzzydate, timestamp_type ts_type)
 {
   int unused;
-  if (check_date(ltime, fuzzy_date, &unused))
+  if (check_date(ltime, fuzzydate, &unused))
   {
     ErrConvTime str(ltime);
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, ts_type, 0, 0);
+    make_truncated_value_warning(thd, Sql_condition::WARN_LEVEL_WARN,
+                                 &str, ts_type, nullptr, nullptr, nullptr);
     return true;
   }
   return false;
@@ -305,7 +313,7 @@ check_date_with_warn(const MYSQL_TIME *ltime, ulonglong fuzzy_date,
 
 
 bool
-adjust_time_range_with_warn(MYSQL_TIME *ltime, uint dec)
+adjust_time_range_with_warn(THD *thd, MYSQL_TIME *ltime, uint dec)
 {
   MYSQL_TIME copy= *ltime;
   ErrConvTime str(&copy);
@@ -313,8 +321,7 @@ adjust_time_range_with_warn(MYSQL_TIME *ltime, uint dec)
   if (check_time_range(ltime, dec, &warnings))
     return true;
   if (warnings)
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, MYSQL_TIMESTAMP_TIME, 0, NullS);
+    thd->push_warning_truncated_wrong_value("time", str.ptr());
   return false;
 }
 
@@ -339,9 +346,9 @@ to_ascii(CHARSET_INFO *cs,
   const char *srcend= src + src_length;
   char *dst0= dst, *dstend= dst + dst_length - 1;
   while (dst < dstend &&
-         (cnvres= (cs->cset->mb_wc)(cs, &wc,
-                                    (const uchar*) src,
-                                    (const uchar*) srcend)) > 0 &&
+         (cnvres= cs->mb_wc(&wc,
+                            (const uchar*) src,
+                            (const uchar*) srcend)) > 0 &&
          wc < 128)
   {
     src+= cnvres;
@@ -352,33 +359,70 @@ to_ascii(CHARSET_INFO *cs,
 }
 
 
-/* Character set-aware version of str_to_time() */
-bool
-str_to_time(CHARSET_INFO *cs, const char *str, size_t length,
-            MYSQL_TIME *l_time, ulonglong fuzzydate, MYSQL_TIME_STATUS *status)
+class TemporalAsciiBuffer: public LEX_CSTRING
 {
   char cnv[32];
-  if ((cs->state & MY_CS_NONASCII) != 0)
+public:
+  TemporalAsciiBuffer(const char *str, size_t length, CHARSET_INFO *cs)
   {
-    length= to_ascii(cs, str, length, cnv, sizeof(cnv));
-    str= cnv;
+    if ((cs->state & MY_CS_NONASCII) != 0)
+    {
+      LEX_CSTRING::str= cnv;
+      LEX_CSTRING::length= to_ascii(cs, str, length, cnv, sizeof(cnv));
+    }
+    else
+    {
+      LEX_CSTRING::str= str;
+      LEX_CSTRING::length= length;
+    }
   }
-  return str_to_time(str, length, l_time, fuzzydate, status);
+};
+
+
+/* Character set-aware version of ascii_to_datetime_or_date_or_time() */
+bool Temporal::str_to_datetime_or_date_or_time(THD *thd, MYSQL_TIME_STATUS *st,
+                                               const char *str, size_t length,
+                                               CHARSET_INFO *cs,
+                                               date_mode_t fuzzydate)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_datetime_or_date_or_time(st, tmp.str, tmp.length, fuzzydate)||
+         add_nanoseconds(thd, &st->warnings, fuzzydate, st->nanoseconds);
 }
 
 
-/* Character set-aware version of str_to_datetime() */
-bool str_to_datetime(CHARSET_INFO *cs, const char *str, size_t length,
-                     MYSQL_TIME *l_time, ulonglong flags,
-                     MYSQL_TIME_STATUS *status)
+/* Character set-aware version of str_to_datetime_or_date() */
+bool Temporal::str_to_datetime_or_date(THD *thd, MYSQL_TIME_STATUS *status,
+                                       const char *str, size_t length,
+                                       CHARSET_INFO *cs,
+                                       date_mode_t flags)
 {
-  char cnv[32];
-  if ((cs->state & MY_CS_NONASCII) != 0)
-  {
-    length= to_ascii(cs, str, length, cnv, sizeof(cnv));
-    str= cnv;
-  }
-  return str_to_datetime(str, length, l_time, flags, status);
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_datetime_or_date(status, tmp.str, tmp.length, flags) ||
+         add_nanoseconds(thd, &status->warnings, flags, status->nanoseconds);
+}
+
+
+/* Character set-aware version of ascii_to_temporal() */
+bool Temporal::str_to_temporal(THD *thd, MYSQL_TIME_STATUS *status,
+                               const char *str, size_t length, CHARSET_INFO *cs,
+                               date_mode_t flags)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  return ascii_to_temporal(status, tmp.str, tmp.length, flags) ||
+         add_nanoseconds(thd, &status->warnings, flags, status->nanoseconds);
+}
+
+
+/* Character set-aware version of str_to_DDhhmmssff() */
+bool Interval_DDhhmmssff::str_to_DDhhmmssff(MYSQL_TIME_STATUS *status,
+                                            const char *str, size_t length,
+                                            CHARSET_INFO *cs, ulong max_hour)
+{
+  TemporalAsciiBuffer tmp(str, length, cs);
+  bool rc= ::str_to_DDhhmmssff(tmp.str, tmp.length, this, UINT_MAX32, status);
+  DBUG_ASSERT(status->warnings || !rc);
+  return rc;
 }
 
 
@@ -387,125 +431,59 @@ bool str_to_datetime(CHARSET_INFO *cs, const char *str, size_t length,
   if string was truncated during conversion.
 
   NOTE
-    See description of str_to_datetime() for more information.
+    See description of str_to_datetime_xxx() for more information.
 */
 
 bool
-str_to_datetime_with_warn(CHARSET_INFO *cs,
-                          const char *str, size_t length, MYSQL_TIME *l_time,
-                          ulonglong flags)
+str_to_datetime_with_warn(THD *thd, CHARSET_INFO *cs,
+                          const char *str, size_t length, MYSQL_TIME *to,
+                          date_mode_t mode)
 {
-  MYSQL_TIME_STATUS status;
-  THD *thd= current_thd;
-  bool ret_val= str_to_datetime(cs, str, length, l_time, flags, &status);
-  if (ret_val || status.warnings)
-    make_truncated_value_warning(thd,
-                                 ret_val ? Sql_condition::WARN_LEVEL_WARN :
-                                 Sql_condition::time_warn_level(status.warnings),
-                                 str, length, flags & TIME_TIME_ONLY ?
-                                 MYSQL_TIMESTAMP_TIME : l_time->time_type, 0, NullS);
-  DBUG_EXECUTE_IF("str_to_datetime_warn",
-                  push_warning(thd, Sql_condition::WARN_LEVEL_NOTE,
-                               ER_YES, str););
-  return ret_val;
+  Temporal::Warn_push warn(thd, nullptr, nullptr, nullptr, to, mode);
+  Temporal_hybrid *t= new(to) Temporal_hybrid(thd, &warn, str, length, cs, mode);
+  return !t->is_valid_temporal();
 }
 
 
-/**
-  converts a pair of numbers (integer part, microseconds) to MYSQL_TIME
-
-  @param neg           sign of the time value
-  @param nr            integer part of the number to convert
-  @param sec_part      microsecond part of the number
-  @param ltime         converted value will be written here
-  @param fuzzydate     conversion flags (TIME_INVALID_DATE, etc)
-  @param str           original number, as an ErrConv. For the warning
-  @param field_name    field name or NULL if not a field. For the warning
-  
-  @returns 0 for success, 1 for a failure
-*/
-static bool number_to_time_with_warn(bool neg, ulonglong nr, ulong sec_part,
-                                     MYSQL_TIME *ltime, ulonglong fuzzydate,
-                                     const ErrConv *str,
-                                     const TABLE_SHARE *s, const char *field_name)
-{
-  int was_cut;
-  longlong res;
-  enum_mysql_timestamp_type ts_type;
-  bool have_warnings;
-
-  if (fuzzydate & TIME_TIME_ONLY)
-  {
-    fuzzydate= TIME_TIME_ONLY; // clear other flags
-    ts_type= MYSQL_TIMESTAMP_TIME;
-    res= number_to_time(neg, nr, sec_part, ltime, &was_cut);
-    have_warnings= MYSQL_TIME_WARN_HAVE_WARNINGS(was_cut);
-  }
-  else
-  {
-    ts_type= MYSQL_TIMESTAMP_DATETIME;
-    if (neg)
-    {
-      res= -1;
-    }
-    else
-    {
-      res= number_to_datetime(nr, sec_part, ltime, fuzzydate, &was_cut);
-      have_warnings= was_cut && (fuzzydate & TIME_NO_ZERO_IN_DATE);
-    }
-  }
-
-  if (res < 0 || have_warnings)
-  {
-    make_truncated_value_warning(current_thd,
-                                 Sql_condition::WARN_LEVEL_WARN, str,
-                                 res < 0 ? MYSQL_TIMESTAMP_ERROR : ts_type,
-                                 s, field_name);
-  }
-  return res < 0;
-}
-
-
-bool double_to_datetime_with_warn(double value, MYSQL_TIME *ltime,
-                                  ulonglong fuzzydate,
+bool double_to_datetime_with_warn(THD *thd, double value, MYSQL_TIME *ltime,
+                                  date_mode_t fuzzydate,
                                   const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvDouble str(value);
-  bool neg= value < 0;
-
-  if (neg)
-    value= -value;
-
-  if (value > LONGLONG_MAX)
-    value= static_cast<double>(LONGLONG_MAX);
-
-  longlong nr= static_cast<ulonglong>(floor(value));
-  uint sec_part= static_cast<ulong>((value - floor(value))*TIME_SECOND_PART_FACTOR);
-  return number_to_time_with_warn(neg, nr, sec_part, ltime, fuzzydate, &str,
-                                  s, field_name);
+  Temporal::Warn_push warn(thd, s ? s->db.str : nullptr,
+                           s ? s->table_name.str : nullptr,
+                           field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, value, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
-bool decimal_to_datetime_with_warn(const my_decimal *value, MYSQL_TIME *ltime,
-                                   ulonglong fuzzydate,
+bool decimal_to_datetime_with_warn(THD *thd, const my_decimal *value,
+                                   MYSQL_TIME *ltime,
+                                   date_mode_t fuzzydate,
                                    const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvDecimal str(value);
-  ulonglong nr;
-  ulong sec_part;
-  bool neg= my_decimal2seconds(value, &nr, &sec_part);
-  return number_to_time_with_warn(neg, nr, sec_part, ltime, fuzzydate, &str,
-                                  s, field_name);
+  Temporal::Warn_push warn(thd, s ? s->db.str : nullptr,
+                           s ? s->table_name.str : nullptr,
+                           field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, value, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
-bool int_to_datetime_with_warn(bool neg, ulonglong value, MYSQL_TIME *ltime,
-                               ulonglong fuzzydate,
+bool int_to_datetime_with_warn(THD *thd, const Longlong_hybrid &nr,
+                               MYSQL_TIME *ltime,
+                               date_mode_t fuzzydate,
                                const TABLE_SHARE *s, const char *field_name)
 {
-  const ErrConvInteger str(neg ? - (longlong) value : (longlong) value, !neg);
-  return number_to_time_with_warn(neg, value, 0, ltime,
-                                  fuzzydate, &str, s, field_name);
+  /*
+    Note: conversion from an integer to TIME can overflow to '838:59:59.999999',
+    so the conversion result can have fractional digits.
+  */
+  Temporal::Warn_push warn(thd, s ? s->db.str : nullptr,
+                           s ? s->table_name.str : nullptr,
+                           field_name, ltime, fuzzydate);
+  Temporal_hybrid *t= new (ltime) Temporal_hybrid(thd, &warn, nr, fuzzydate);
+  return !t->is_valid_temporal();
 }
 
 
@@ -530,7 +508,7 @@ bool int_to_datetime_with_warn(bool neg, ulonglong value, MYSQL_TIME *ltime,
 
 my_time_t TIME_to_timestamp(THD *thd, const MYSQL_TIME *t, uint *error_code)
 {
-  thd->time_zone_used= 1;
+  thd->used|= THD::TIME_ZONE_USED;
   return thd->variables.time_zone->TIME_to_gmt_sec(t, error_code);
 }
 
@@ -552,7 +530,7 @@ void localtime_to_TIME(MYSQL_TIME *to, struct tm *from)
 }
 
 
-void calc_time_from_sec(MYSQL_TIME *to, long seconds, long microseconds)
+void calc_time_from_sec(MYSQL_TIME *to, ulong seconds, ulong microseconds)
 {
   long t_seconds;
   // to->neg is not cleared, it may already be set to a useful value
@@ -568,323 +546,6 @@ void calc_time_from_sec(MYSQL_TIME *to, long seconds, long microseconds)
 }
 
 
-/*
-  Parse a format string specification
-
-  SYNOPSIS
-    parse_date_time_format()
-    format_type		Format of string (time, date or datetime)
-    format_str		String to parse
-    format_length	Length of string
-    date_time_format	Format to fill in
-
-  NOTES
-    Fills in date_time_format->positions for all date time parts.
-
-    positions marks the position for a datetime element in the format string.
-    The position array elements are in the following order:
-    YYYY-DD-MM HH-MM-DD.FFFFFF AM
-    0    1  2  3  4  5  6      7
-
-    If positions[0]= 5, it means that year will be the forth element to
-    read from the parsed date string.
-
-  RETURN
-    0	ok
-    1	error
-*/
-
-bool parse_date_time_format(timestamp_type format_type, 
-			    const char *format, uint format_length,
-			    DATE_TIME_FORMAT *date_time_format)
-{
-  uint offset= 0, separators= 0;
-  const char *ptr= format, *format_str;
-  const char *end= ptr+format_length;
-  uchar *dt_pos= date_time_format->positions;
-  /* need_p is set if we are using AM/PM format */
-  bool need_p= 0, allow_separator= 0;
-  ulong part_map= 0, separator_map= 0;
-  const char *parts[16];
-
-  date_time_format->time_separator= 0;
-  date_time_format->flag= 0;			// For future
-
-  /*
-    Fill position with 'dummy' arguments to found out if a format tag is
-    used twice (This limit's the format to 255 characters, but this is ok)
-  */
-  dt_pos[0]= dt_pos[1]= dt_pos[2]= dt_pos[3]=
-    dt_pos[4]= dt_pos[5]= dt_pos[6]= dt_pos[7]= 255;
-
-  for (; ptr != end; ptr++)
-  {
-    if (*ptr == '%' && ptr+1 != end)
-    {
-      uint UNINIT_VAR(position);
-      switch (*++ptr) {
-      case 'y':					// Year
-      case 'Y':
-	position= 0;
-	break;
-      case 'c':					// Month
-      case 'm':
-	position= 1;
-	break;
-      case 'd':
-      case 'e':
-	position= 2;
-	break;
-      case 'h':
-      case 'I':
-      case 'l':
-	need_p= 1;				// Need AM/PM
-	/* Fall through */
-      case 'k':
-      case 'H':
-	position= 3;
-	break;
-      case 'i':
-	position= 4;
-	break;
-      case 's':
-      case 'S':
-	position= 5;
-	break;
-      case 'f':
-	position= 6;
-	if (dt_pos[5] != offset-1 || ptr[-2] != '.')
-	  return 1;				// Wrong usage of %f
-	break;
-      case 'p':					// AM/PM
-	if (offset == 0)			// Can't be first
-	  return 0;
-	position= 7;
-	break;
-      default:
-	return 1;				// Unknown controll char
-      }
-      if (dt_pos[position] != 255)		// Don't allow same tag twice
-	return 1;
-      parts[position]= ptr-1;
-
-      /*
-	If switching from time to date, ensure that all time parts
-	are used
-      */
-      if (part_map && position <= 2 && !(part_map & (1 | 2 | 4)))
-	offset=5;
-      part_map|= (ulong) 1 << position;
-      dt_pos[position]= offset++;
-      allow_separator= 1;
-    }
-    else
-    {
-      /*
-	Don't allow any characters in format as this could easily confuse
-	the date reader
-      */
-      if (!allow_separator)
-	return 1;				// No separator here
-      allow_separator= 0;			// Don't allow two separators
-      separators++;
-      /* Store in separator_map which parts are punct characters */
-      if (my_ispunct(&my_charset_latin1, *ptr))
-	separator_map|= (ulong) 1 << (offset-1);
-      else if (!my_isspace(&my_charset_latin1, *ptr))
-	return 1;
-    }
-  }
-
-  /* If no %f, specify it after seconds.  Move %p up, if necessary */
-  if ((part_map & 32) && !(part_map & 64))
-  {
-    dt_pos[6]= dt_pos[5] +1;
-    parts[6]= parts[5];				// For later test in (need_p)
-    if (dt_pos[6] == dt_pos[7])			// Move %p one step up if used
-      dt_pos[7]++;
-  }
-
-  /*
-    Check that we have not used a non legal format specifier and that all
-    format specifiers have been used
-
-    The last test is to ensure that %p is used if and only if
-    it's needed.
-  */
-  if ((format_type == MYSQL_TIMESTAMP_DATETIME &&
-       !test_all_bits(part_map, (1 | 2 | 4 | 8 | 16 | 32))) ||
-      (format_type == MYSQL_TIMESTAMP_DATE && part_map != (1 | 2 | 4)) ||
-      (format_type == MYSQL_TIMESTAMP_TIME &&
-       !test_all_bits(part_map, 8 | 16 | 32)) ||
-      !allow_separator ||			// %option should be last
-      (need_p && dt_pos[6] +1 != dt_pos[7]) ||
-      (need_p ^ (dt_pos[7] != 255)))
-    return 1;
-
-  if (dt_pos[6] != 255)				// If fractional seconds
-  {
-    /* remove fractional seconds from later tests */
-    uint pos= dt_pos[6] -1;
-    /* Remove separator before %f from sep map */
-    separator_map= ((separator_map & ((ulong) (1 << pos)-1)) |
-		    ((separator_map & ~((ulong) (1 << pos)-1)) >> 1));
-    if (part_map & 64)			      
-    {
-      separators--;				// There is always a separator
-      need_p= 1;				// force use of separators
-    }
-  }
-
-  /*
-    Remove possible separator before %p from sep_map
-    (This can either be at position 3, 4, 6 or 7) h.m.d.%f %p
-  */
-  if (dt_pos[7] != 255)
-  {
-    if (need_p && parts[7] != parts[6]+2)
-      separators--;
-  }     
-  /*
-    Calculate if %p is in first or last part of the datetime field
-
-    At this point we have either %H-%i-%s %p 'year parts' or
-    'year parts' &H-%i-%s %p" as %f was removed above
-  */
-  offset= dt_pos[6] <= 3 ? 3 : 6;
-  /* Remove separator before %p from sep map */
-  separator_map= ((separator_map & ((ulong) (1 << offset)-1)) |
-		  ((separator_map & ~((ulong) (1 << offset)-1)) >> 1));
-
-  format_str= 0;
-  switch (format_type) {
-  case MYSQL_TIMESTAMP_DATE:
-    format_str= known_date_time_formats[INTERNAL_FORMAT].date_format;
-    /* fall through */
-  case MYSQL_TIMESTAMP_TIME:
-    if (!format_str)
-      format_str=known_date_time_formats[INTERNAL_FORMAT].time_format;
-
-    /*
-      If there is no separators, allow the internal format as we can read
-      this.  If separators are used, they must be between each part
-    */
-    if (format_length == 6 && !need_p &&
-	!my_strnncoll(&my_charset_bin,
-		      (const uchar *) format, 6, 
-		      (const uchar *) format_str, 6))
-      return 0;
-    if (separator_map == (1 | 2))
-    {
-      if (format_type == MYSQL_TIMESTAMP_TIME)
-      {
-	if (*(format+2) != *(format+5))
-	  break;				// Error
-	/* Store the character used for time formats */
-	date_time_format->time_separator= *(format+2);
-      }
-      return 0;
-    }
-    break;
-  case MYSQL_TIMESTAMP_DATETIME:
-    /*
-      If there is no separators, allow the internal format as we can read
-      this.  If separators are used, they must be between each part.
-      Between DATE and TIME we also allow space as separator
-    */
-    if ((format_length == 12 && !need_p &&
-	 !my_strnncoll(&my_charset_bin, 
-		       (const uchar *) format, 12,
-		       (const uchar*) known_date_time_formats[INTERNAL_FORMAT].datetime_format,
-		       12)) ||
-	(separators == 5 && separator_map == (1 | 2 | 8 | 16)))
-      return 0;
-    break;
-  default:
-    DBUG_ASSERT(0);
-    break;
-  }
-  return 1;					// Error
-}
-
-
-/*
-  Create a DATE_TIME_FORMAT object from a format string specification
-
-  SYNOPSIS
-    date_time_format_make()
-    format_type		Format to parse (time, date or datetime)
-    format_str		String to parse
-    format_length	Length of string
-
-  NOTES
-    The returned object should be freed with my_free()
-
-  RETURN
-    NULL ponter:	Error
-    new object
-*/
-
-DATE_TIME_FORMAT
-*date_time_format_make(timestamp_type format_type,
-		       const char *format_str, uint format_length)
-{
-  DATE_TIME_FORMAT tmp;
-
-  if (format_length && format_length < 255 &&
-      !parse_date_time_format(format_type, format_str,
-			      format_length, &tmp))
-  {
-    tmp.format.str=    format_str;
-    tmp.format.length= format_length;
-    return date_time_format_copy((THD *)0, &tmp);
-  }
-  return 0;
-}
-
-
-/*
-  Create a copy of a DATE_TIME_FORMAT object
-
-  SYNOPSIS
-    date_and_time_format_copy()
-    thd			Set if variable should be allocated in thread mem
-    format		format to copy
-
-  NOTES
-    The returned object should be freed with my_free()
-
-  RETURN
-    NULL ponter:	Error
-    new object
-*/
-
-DATE_TIME_FORMAT *date_time_format_copy(THD *thd, DATE_TIME_FORMAT *format)
-{
-  DATE_TIME_FORMAT *new_format;
-  size_t length= sizeof(*format) + format->format.length + 1;
-  char *format_pos;
-
-  if (thd)
-    new_format= (DATE_TIME_FORMAT *) thd->alloc(length);
-  else
-    new_format=  (DATE_TIME_FORMAT *) my_malloc(length, MYF(MY_WME));
-  if (new_format)
-  {
-    /* Put format string after current pos */
-    new_format->format.str= format_pos= (char*) (new_format+1);
-    memcpy((char*) new_format->positions, (char*) format->positions,
-	   sizeof(format->positions));
-    new_format->time_separator= format->time_separator;
-    /* We make the string null terminated for easy printf in SHOW VARIABLES */
-    memcpy(format_pos, format->format.str, format->format.length);
-    format_pos[format->format.length]= 0;
-    new_format->format.length= format->format.length;
-  }
-  return new_format;
-}
-
-
 KNOWN_DATE_TIME_FORMAT known_date_time_formats[6]=
 {
   {"USA", "%m.%d.%Y", "%Y-%m-%d %H.%i.%s", "%h:%i:%s %p" },
@@ -933,52 +594,13 @@ void make_truncated_value_warning(THD *thd,
                                   Sql_condition::enum_warning_level level,
                                   const ErrConv *sval,
 				  timestamp_type time_type,
-                                  const TABLE_SHARE *s, const char *field_name)
+                                  const char *db_name, const char *table_name,
+                                  const char *field_name)
 {
-  char warn_buff[MYSQL_ERRMSG_SIZE];
-  const char *type_str;
-  CHARSET_INFO *cs= &my_charset_latin1;
-
-  switch (time_type) {
-    case MYSQL_TIMESTAMP_DATE: 
-      type_str= "date";
-      break;
-    case MYSQL_TIMESTAMP_TIME:
-      type_str= "time";
-      break;
-    case MYSQL_TIMESTAMP_DATETIME:  // FALLTHROUGH
-    default:
-      type_str= "datetime";
-      break;
-  }
-  if (field_name)
-  {
-    const char *db_name= s->db.str;
-    const char *table_name= s->table_name.str;
-
-    if (!db_name)
-      db_name= "";
-    if (!table_name)
-      table_name= "";
-
-    cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                       ER_THD(thd, ER_TRUNCATED_WRONG_VALUE_FOR_FIELD),
-                       type_str, sval->ptr(),
-                       db_name, table_name, field_name,
-                       (ulong) thd->get_stmt_da()->current_row_for_warning());
-  }
-  else
-  {
-    if (time_type > MYSQL_TIMESTAMP_ERROR)
-      cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                         ER_THD(thd, ER_TRUNCATED_WRONG_VALUE),
-                         type_str, sval->ptr());
-    else
-      cs->cset->snprintf(cs, warn_buff, sizeof(warn_buff),
-                         ER_THD(thd, ER_WRONG_VALUE), type_str, sval->ptr());
-  }
-  push_warning(thd, level,
-               ER_TRUNCATED_WRONG_VALUE, warn_buff);
+  const char *type_str= Temporal::type_name_by_timestamp_type(time_type);
+  return thd->push_warning_wrong_or_truncated_value
+    (level, time_type <= MYSQL_TIMESTAMP_ERROR, type_str, sval->ptr(),
+     db_name, table_name, field_name);
 }
 
 
@@ -989,10 +611,10 @@ void make_truncated_value_warning(THD *thd,
                    (X)->second_part)
 #define GET_PART(X, N) X % N ## LL; X/= N ## LL
 
-bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
-                       const INTERVAL &interval)
+bool date_add_interval(THD *thd, MYSQL_TIME *ltime, interval_type int_type,
+                       const INTERVAL &interval, bool push_warn)
 {
-  long period, sign;
+  long sign;
 
   sign= (interval.neg == (bool)ltime->neg ? 1 : -1);
 
@@ -1063,13 +685,17 @@ bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
     break;
   }
   case INTERVAL_WEEK:
-    period= (calc_daynr(ltime->year,ltime->month,ltime->day) +
-             sign * (long) interval.day);
+  {
+    longlong period= calc_daynr(ltime->year, ltime->month, ltime->day) +
+                     (longlong) sign * (longlong) interval.day;
+    if (period < 0 || period > 0x7FFFFFFF)
+      goto invalid_date;
     /* Daynumber from year 0 to 9999-12-31 */
     if (get_date_from_daynr((long) period,&ltime->year,&ltime->month,
                             &ltime->day))
       goto invalid_date;
     break;
+  }
   case INTERVAL_YEAR:
     ltime->year+= sign * (long) interval.year;
     if ((ulong) ltime->year >= 10000L)
@@ -1081,8 +707,9 @@ bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
   case INTERVAL_YEAR_MONTH:
   case INTERVAL_QUARTER:
   case INTERVAL_MONTH:
-    period= (ltime->year*12 + sign * (long) interval.year*12 +
-	     ltime->month-1 + sign * (long) interval.month);
+  {
+    long period= (ltime->year*12 + sign * (long) interval.year*12 +
+	         ltime->month-1 + sign * (long) interval.month);
     if ((ulong) period >= 120000L)
       goto invalid_date;
     ltime->year= (uint) (period / 12);
@@ -1095,6 +722,7 @@ bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
 	ltime->day++;				// Leap-year
     }
     break;
+  }
   default:
     goto null_date;
   }
@@ -1103,8 +731,8 @@ bool date_add_interval(MYSQL_TIME *ltime, interval_type int_type,
     return 0;                                   // Ok
 
 invalid_date:
+  if (push_warn)
   {
-    THD *thd= current_thd;
     push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                         ER_DATETIME_FUNCTION_OVERFLOW,
                         ER_THD(thd, ER_DATETIME_FUNCTION_OVERFLOW),
@@ -1144,7 +772,7 @@ null_date:
 
 bool
 calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
-               int l_sign, longlong *seconds_out, long *microseconds_out)
+               int l_sign, ulonglong *seconds_out, ulong *microseconds_out)
 {
   long days;
   bool neg;
@@ -1171,10 +799,10 @@ calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
   }
 
   microseconds= ((longlong)days * SECONDS_IN_24H +
-                 (longlong)(l_time1->hour*3600L +
+                 (longlong)(l_time1->hour*3600LL +
                             l_time1->minute*60L +
                             l_time1->second) -
-                 l_sign*(longlong)(l_time2->hour*3600L +
+                 l_sign*(longlong)(l_time2->hour*3600LL +
                                    l_time2->minute*60L +
                                    l_time2->second)) * 1000000LL +
                 (longlong)l_time1->second_part -
@@ -1186,17 +814,17 @@ calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
     microseconds= -microseconds;
     neg= 1;
   }
-  *seconds_out= microseconds/1000000L;
-  *microseconds_out= (long) (microseconds%1000000L);
+  *seconds_out= (ulonglong) microseconds/1000000L;
+  *microseconds_out= (ulong) (microseconds%1000000L);
   return neg;
 }
 
 
 bool calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
-                    int l_sign, MYSQL_TIME *l_time3, ulonglong fuzzydate)
+                    int l_sign, MYSQL_TIME *l_time3, date_mode_t fuzzydate)
 {
-  longlong seconds;
-  long microseconds;
+  ulonglong seconds;
+  ulong microseconds;
   bzero((char *) l_time3, sizeof(*l_time3));
   l_time3->neg= calc_time_diff(l_time1, l_time2, l_sign,
 			       &seconds, &microseconds);
@@ -1215,7 +843,7 @@ bool calc_time_diff(const MYSQL_TIME *l_time1, const MYSQL_TIME *l_time2,
     ("invalid" means > TIME_MAX_SECOND)
   */
   set_if_smaller(seconds, INT_MAX32);
-  calc_time_from_sec(l_time3, (long) seconds, microseconds);
+  calc_time_from_sec(l_time3, (ulong) seconds, microseconds);
   return ((fuzzydate & TIME_NO_ZERO_DATE) && (seconds == 0) &&
           (microseconds == 0));
 }
@@ -1272,56 +900,6 @@ bool time_to_datetime(MYSQL_TIME *ltime)
 }
 
 
-/**
-  Return a valid DATE or DATETIME value from an arbitrary MYSQL_TIME.
-  If ltime is TIME, it's first converted to DATETIME.
-  If ts_type is DATE, hhmmss is set to zero.
-  The date part of the result is checked against fuzzy_date.
-
-  @param   ltime       The value to convert.
-  @param   fuzzy_date  Flags to check date.
-  @param   ts_type     The type to convert to.
-  @return  false on success, true of error (negative time).*/
-bool
-make_date_with_warn(MYSQL_TIME *ltime, ulonglong fuzzy_date,
-                    timestamp_type ts_type)
-{
-  DBUG_ASSERT(ts_type == MYSQL_TIMESTAMP_DATE ||
-              ts_type == MYSQL_TIMESTAMP_DATETIME);
-  if (ltime->time_type == MYSQL_TIMESTAMP_TIME && time_to_datetime(ltime))
-  {
-    /* e.g. negative time */
-    ErrConvTime str(ltime);
-    make_truncated_value_warning(current_thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, ts_type, 0, 0);
-    return true;
-  }
-  if ((ltime->time_type= ts_type) == MYSQL_TIMESTAMP_DATE)
-    ltime->hour= ltime->minute= ltime->second= ltime->second_part= 0;
-  return check_date_with_warn(ltime, fuzzy_date, ts_type);
-}
-
-
-/*
-  Convert a TIME value to DAY-TIME interval, e.g. for extraction:
-    EXTRACT(DAY FROM x), EXTRACT(HOUR FROM x), etc.
-  Moves full days from ltime->hour to ltime->day.
-  Note, time_type is set to MYSQL_TIMESTAMP_NONE, to make sure that
-  the structure is not used for anything else other than extraction:
-  non-extraction TIME functions expect zero day value!
-*/
-void time_to_daytime_interval(MYSQL_TIME *ltime)
-{
-  DBUG_ASSERT(ltime->time_type == MYSQL_TIMESTAMP_TIME);
-  DBUG_ASSERT(ltime->year == 0);
-  DBUG_ASSERT(ltime->month == 0);
-  DBUG_ASSERT(ltime->day == 0);
-  ltime->day= ltime->hour / 24;
-  ltime->hour%= 24;
-  ltime->time_type= MYSQL_TIMESTAMP_NONE;
-}
-
-
 /*** Conversion from TIME to DATETIME ***/
 
 /*
@@ -1349,8 +927,8 @@ mix_date_and_time_complex(MYSQL_TIME *ldate, const MYSQL_TIME *ltime)
 {
   DBUG_ASSERT(ldate->time_type == MYSQL_TIMESTAMP_DATE ||
               ldate->time_type == MYSQL_TIMESTAMP_DATETIME);
-  longlong seconds;
-  long days, useconds;
+  ulonglong seconds;
+  ulong days, useconds;
   int sign= ltime->neg ? 1 : -1;
   ldate->neg= calc_time_diff(ldate, ltime, sign, &seconds, &useconds);
 
@@ -1386,7 +964,7 @@ mix_date_and_time(MYSQL_TIME *to, const MYSQL_TIME *from)
 void set_current_date(THD *thd, MYSQL_TIME *to)
 {
   thd->variables.time_zone->gmt_sec_to_TIME(to, thd->query_start());
-  thd->time_zone_used= 1;
+  thd->used|= THD::TIME_ZONE_USED;
   datetime_to_date(to);
 }
 
@@ -1440,7 +1018,7 @@ time_to_datetime(THD *thd, const MYSQL_TIME *from, MYSQL_TIME *to)
 bool
 time_to_datetime_with_warn(THD *thd,
                            const MYSQL_TIME *from, MYSQL_TIME *to,
-                           ulonglong fuzzydate)
+                           date_conv_mode_t fuzzydate)
 {
   int warn= 0;
   DBUG_ASSERT(from->time_type == MYSQL_TIMESTAMP_TIME);
@@ -1456,34 +1034,13 @@ time_to_datetime_with_warn(THD *thd,
         check_date(to, fuzzydate, &warn)))
   {
     ErrConvTime str(from);
-    make_truncated_value_warning(thd, Sql_condition::WARN_LEVEL_WARN,
-                                 &str, MYSQL_TIMESTAMP_DATETIME, 0, 0); 
+    thd->push_warning_truncated_wrong_value("datetime", str.ptr());
     return true;
   }
   return false;
 }
 
 
-bool datetime_to_time_with_warn(THD *thd, const MYSQL_TIME *dt,
-                                MYSQL_TIME *tm, uint dec)
-{
-  if (thd->variables.old_behavior & OLD_MODE_ZERO_DATE_TIME_CAST)
-  {
-    *tm= *dt;
-    datetime_to_time(tm);
-    return false;
-  }
-  else /* new mode */
-  {
-    MYSQL_TIME current_date;
-    set_current_date(thd, &current_date);
-    calc_time_diff(dt, &current_date, 1, tm, 0);
-  }
-  int warnings= 0;
-  return check_time_range(tm, dec, &warnings);
-}
-
-
 longlong pack_time(const MYSQL_TIME *my_time)
 {
   return  ((((((my_time->year     * 13ULL +
